import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:dart_jsonwebtoken/dart_jsonwebtoken.dart';
import 'package:dart_nuntio_cloud/dart_nuntio_cloud.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart';
import 'package:geocoding/geocoding.dart';
import 'package:grpc/src/client/transport/http2_credentials.dart'
    show ChannelCredentials;
import 'package:jwt_decoder/jwt_decoder.dart';
import 'package:nuntio_hera/hera/helpers/decode_jwt.dart';
import 'package:nuntio_hera/hera/models/find_options.dart';
import 'package:nuntio_hera/hera/models/user_options.dart';
import 'package:nuntio_hera/hera/hera_extension.dart';

import 'package:uni_links/uni_links.dart';
import 'package:url_launcher/url_launcher.dart';
import '../hera.pbgrpc.dart';
import 'channel/hera_channel.dart'
    if (dart.library.html) 'channel/hera_channel_web.dart';

const String _currentUserKey = "hera-current-user";
const String _accessTokenKey = "hera-access-token";
const String _refreshTokenKey = "hera-refresh-token";

class Hera {
  // userClient is used to make requests
  static late ServiceClient _heraClient;
  static Function? onLogout;
  static late String _namespace;
  static late bool _debug;
  static late TransportCredentials _transportCredentials;
  static late FlutterSecureStorage _secureStorage;
  static String _publicApiKey = "";
  static String _refreshToken = "";
  static String _jwtPublicKey = "";
  static User _currentUser = User();
  static StreamSubscription? _sub;
  static Placemark? _placemark;
  static late String _oidcRedirectAddress;
  static late String _oidcListenFor;
  static late String _apiUrl;

  static Future<void> initialize({
    String? apiUrl,
    String? publicApiKey,
    String? namespace,
    bool? debug,
    TransportCredentials? credentials,
    String? oidcRedirectAddress,
    String? oidcListenFor,
  }) async {
    _oidcRedirectAddress = oidcRedirectAddress ??
        (kIsWeb ? "" : "com.example.nuntioapp:/oauth2Callback");
    _oidcListenFor = oidcListenFor ?? (kIsWeb ? "" : "com.example.nuntioapp");
    // set values
    _namespace = namespace ?? "";
    _debug = debug ?? false;
    // get api url
    _publicApiKey = publicApiKey ?? "";
    if (apiUrl == null) {
      var claims = parseJwt(_publicApiKey);
      if (claims.containsKey("api_url") == false) {
        throw Exception("missing required api url");
      }
      _apiUrl = "https://" + claims["api_url"];
    } else {
      _apiUrl = apiUrl;
    }
    // setup transport credentials
    _transportCredentials =
        credentials ?? CloudCredentials(apiUrl: Uri.parse(_apiUrl));
    // create client
    _heraClient =
        await HeraChannel.userServiceClient(_apiUrl, _transportCredentials);
    // get block user public key
    HeraRequest publicKeysReq = HeraRequest();
    publicKeysReq.cloudToken = _publicApiKey;
    HeraResponse publicKeysResp =
        await _heraClient.getPublicKeys(publicKeysReq);
    _jwtPublicKey = publicKeysResp.publicKeys["hera-public-key"] ?? "";
    // create secure storage
    _secureStorage = FlutterSecureStorage();
    //initUniLinks();
  }

  static Future<User> getCurrentUser() async {
    if (_currentUser.id != "") {
      return _currentUser;
    }
    var jsonCurrentUser = await _secureStorage.read(key: _currentUserKey);
    if (jsonCurrentUser != null && jsonCurrentUser != "") {
      _currentUser = User.fromJson(jsonCurrentUser);
      return _currentUser;
    }
    return User();
  }

  static void _setCurrentUser(User currentUser) async {
    _currentUser = currentUser;
    await _secureStorage.write(
        key: _currentUserKey, value: currentUser.writeToJson());
  }

  static Future<String> getAccessToken() async {
    if (_publicApiKey != "") {
      return _publicApiKey;
    }
    var accessToken = await _secureStorage.read(key: _accessTokenKey);
    if (accessToken != "") {
      _publicApiKey = accessToken!;
      return accessToken;
    }
    return "";
  }

  static void _setAccessToken(String accessToken) async {
    _publicApiKey = accessToken;
    await _secureStorage.write(key: _accessTokenKey, value: accessToken);
  }

  static Future<String> _getRefreshToken() async {
    if (_refreshToken != "") {
      return _refreshToken;
    }
    var refreshToken = await _secureStorage.read(key: _refreshTokenKey);
    if (refreshToken != "") {
      _refreshToken = refreshToken!;
      return refreshToken;
    }
    return "";
  }

  static void _setRefreshToken(String refreshToken) async {
    _refreshToken = refreshToken;
    await _secureStorage.write(key: _refreshTokenKey, value: refreshToken);
  }

  static Future<User> createUser({
    UserOptions? userOptions,
    String? password,
  }) async {
    HeraRequest req = HeraRequest();
    User user = User();
    if (userOptions != null) {
      user.image = userOptions.image ?? "";
      user.id = userOptions.id ?? "";
      user.username = userOptions.username ?? "";
      user.phone = userOptions.phone ?? "";
      user.email = userOptions.email ?? "";
      user.gender = userOptions.gender ?? "";
      user.firstName = userOptions.firstName ?? "";
      user.lastName = userOptions.lastName ?? "";
      user.ssn = userOptions.ssn ?? "";
    }
    user.password = Hash(body: password ?? "");
    req.cloudToken = _publicApiKey;
    req.user = user;
    req.namespace = _namespace;
    try {
      HeraResponse resp = await _heraClient.create(req);
      return resp.user;
    } catch (e) {
      if (_debug == true) {
        print("could not create user with err: " + e.toString());
      }
      rethrow;
    }
  }

  static Future<LoginSession> login({
    FindOptions? findOptions,
    required String password,
  }) async {
    HeraRequest req = HeraRequest();
    User user = User();
    if (findOptions == null || findOptions.validate() == false) {
      throw Exception(
          "missing required identifier: id, username, phone or email");
    } else {
      user.id = findOptions.id ?? "";
      user.email = findOptions.email ?? "";
      user.username = findOptions.username ?? "";
      user.phone = findOptions.phone ?? "";
    }
    req.user = user;
    String cloudToken = _publicApiKey;
    req.cloudToken = cloudToken;
    user.password = Hash(body: password);
    req.namespace = _namespace;
    // set metadata for token
    Token _token = Token();
    _token.deviceInfo = await HeraExtenstion.getDeviceInfo();
    _placemark = _placemark ?? await HeraExtenstion.determinePlacemark();
    if (_placemark != null) {
      _token.loggedInFrom = _placemark?.locality ?? "";
    }
    req.token = _token;
    try {
      HeraResponse resp = await _heraClient.login(req);
      if (resp.loginSession.loginStatus != LoginStatus.AUTHENTICATED) {
        return resp.loginSession;
      }
      if (resp.token.accessToken == "" || resp.token.refreshToken == "") {
        throw Exception("token is null. Contact info@nuntio.io");
      }
      if (resp.user.id == "") {
        throw Exception("user is null. Contact info@nuntio.io");
      }
      _setCurrentUser(resp.user);
      _setAccessToken(resp.token.accessToken);
      _setRefreshToken(resp.token.refreshToken);
    } catch (e) {
      if (_debug == true) print("could not login user: " + e.toString());
      rethrow;
    }
    return LoginSession()..loginStatus = LoginStatus.AUTHENTICATED;
  }

  static Future<LoginSession> loginWithOidc({
    required OidcProviderType type,
    required String authorizationUrl,
  }) async {
    await HeraExtenstion.redirect(Uri.parse(authorizationUrl));
    final Completer<Uri> completer = Completer<Uri>();
    _sub = linkStream.listen((String? uri) async {
      if (uri.toString().startsWith(_oidcListenFor)) {
        if (Platform.isIOS || Platform.isAndroid || Platform.isFuchsia) {
          await closeInAppWebView();
        }
        completer.complete(Uri.parse(uri!));
      }
    });
    Uri _responseUrl = await completer.future;
    _sub?.cancel();
    String cloudToken = _publicApiKey;
    HeraRequest req = HeraRequest();
    req.cloudToken = cloudToken;
    req.oidcProvider = OidcProvider()..type = type;
    req.redirectAddr = _oidcRedirectAddress;
    _responseUrl.queryParameters["code"];
    req.authorizationCode = _responseUrl.queryParameters["code"] ?? "";
    HeraResponse resp = await _heraClient.continueWith(req);
    if (resp.token.accessToken == "" || resp.token.refreshToken == "") {
      throw Exception("token is null. Contact info@nuntio.io");
    }
    if (resp.user.id == "") {
      throw Exception("user is null. Contact info@nuntio.io");
    }
    _setCurrentUser(resp.user);
    _setAccessToken(resp.token.accessToken);
    _setRefreshToken(resp.token.refreshToken);
    return LoginSession()..loginStatus = LoginStatus.AUTHENTICATED;
  }

  static Future<void> logout() async {
    if (await isAuthenticated() == false) {
      throw Exception("user is currently not logged in");
    }
    try {
      // block current access token
      HeraRequest req = HeraRequest();
      req.cloudToken = _publicApiKey;
      req.tokenPointer = await getAccessToken();
      req.namespace = _namespace;
      _heraClient.blockToken(req);
      // block current refresh token
      req.tokenPointer = await _getRefreshToken();
      _heraClient.blockToken(req);
      // remove from secure storage
      _secureStorage.delete(key: _accessTokenKey);
      _secureStorage.delete(key: _refreshTokenKey);
      _secureStorage.delete(key: _currentUserKey);
      _currentUser = User();
      _publicApiKey = "";
      _refreshToken = "";
      if (onLogout != null) {
        onLogout!();
      }
    } catch (e) {
      if (_debug == true) print("could not logout user: " + e.toString());
      rethrow;
    }
  }

  static Future<void> verifyEmail({
    FindOptions? findOptions,
    required String code,
  }) async {
    User user = User();
    if (findOptions == null || findOptions.validate() == false) {
      throw Exception(
          "missing required identifier: id, username, phone or email");
    } else {
      user.id = findOptions.id ?? "";
      user.email = findOptions.email ?? "";
      user.username = findOptions.username ?? "";
      user.phone = findOptions.phone ?? "";
    }
    if (code == "") {
      throw Exception("email verification code is empty");
    }
    String cloudToken = _publicApiKey;
    HeraRequest req = HeraRequest();
    VerificationCode verificationCode = VerificationCode();
    verificationCode.code = Hash(body: code);
    req.verificationCode = verificationCode;
    req.cloudToken = cloudToken;
    req.user = user;
    req.namespace = _namespace;
    await _heraClient.verifyEmail(req);
  }

  static Future<void> verifyPhone({
    FindOptions? findOptions,
    required String code,
  }) async {
    User user = User();
    if (findOptions == null || findOptions.validate() == false) {
      throw Exception(
          "missing required identifier: id, username, phone or email");
    } else {
      user.id = findOptions.id ?? "";
      user.email = findOptions.email ?? "";
      user.username = findOptions.username ?? "";
      user.phone = findOptions.phone ?? "";
    }
    if (code == "") {
      throw Exception("email verification code is empty");
    }
    String cloudToken = _publicApiKey;
    HeraRequest req = HeraRequest();
    VerificationCode verificationCode = VerificationCode();
    verificationCode.code = Hash(body: code);
    req.verificationCode = verificationCode;
    req.cloudToken = cloudToken;
    req.user = user;
    req.namespace = _namespace;
    await _heraClient.verifyPhone(req);
  }

  static Future<bool> isAuthenticated() async {
    User currentUser = await getCurrentUser();
    if (currentUser.id != "") {
      try {
        var accessToken = await getAccessToken();
        if (accessToken != "" &&
            JwtDecoder.getRemainingTime(accessToken).inMinutes > 2) {
          try {
            if (_jwtPublicKey == "") {
              throw Exception("empty jwt public key");
            }
            JWT.verify(accessToken, RSAPublicKey(_jwtPublicKey));
            return true;
          } catch (e) {
            if (_debug == true) {
              print("could not verify token with err" + e.toString());
            }
          }
        }
        // token is expired - refresh
        HeraRequest req = HeraRequest();
        req.cloudToken = _publicApiKey;
        // set metadata for token
        Token _token = Token();
        _token.deviceInfo = await HeraExtenstion.getDeviceInfo();
        _placemark = _placemark ?? await HeraExtenstion.determinePlacemark();
        if (_placemark != null) {
          _token.loggedInFrom = _placemark?.locality ?? "";
        }
        _token.refreshToken = await _getRefreshToken();
        req.token = _token;
        req.namespace = _namespace;
        HeraResponse refreshResp = await _heraClient.refreshToken(req);
        _setAccessToken(refreshResp.token.accessToken);
        _setRefreshToken(refreshResp.token.refreshToken);
        return true;
      } catch (e) {
        if (_debug == true) {
          print("could not refresh token with err: " + e.toString());
        }
        return false;
      }
    }
    return false;
  }

  static Future<Config> getConfig() async {
    HeraRequest req = HeraRequest();
    req.cloudToken = _publicApiKey;
    req.namespace = _namespace;
    req.redirectAddr = _oidcRedirectAddress;
    try {
      return (await _heraClient.getConfig(req)).config;
    } catch (e) {
      _debug ? print(e) : () {};
      rethrow;
    }
  }
}

class NoTransportCredentials implements TransportCredentials {
  @override
  Future<ChannelCredentials> getTransportCredentials() async {
    return ChannelCredentials.insecure();
  }
}
